;;; Copyright (C) 1992, 1994 R. Kent Dybvig
;;; Permission to copy this software, in whole or in part, to use this
;;; software for any lawful purpose, and to redistribute this software
;;; is granted subject to the restriction that all copies made of this
;;; software must include this copyright notice in full.  This software
;;; is provided AS IS, with NO WARRANTY, EITHER EXPRESS OR IMPLIED,
;;; INCLUDING BUT NOT LIMITED TO IMPLIED WARRANTIES OF MERCHANTABILITY
;;; OR FITNESS FOR ANY PARTICULAR PURPOSE.  IN NO EVENT SHALL THE
;;; AUTHORS BE LIABLE FOR CONSEQUENTIAL OR INCIDENTAL DAMAGES OF ANY
;;; NATURE WHATSOEVER.

"syntax.ss"
(eval-when (compile) (current-expand sc-expand))

(define \#sc-default-get-expander
   (lambda (x)
      (getprop x '*macro-transformer*)))

(define \#sc-get-expander \#sc-default-get-expander)

;;; expand.ss

;;; notes:

;;; "begin" is treated as a splicing construct at top level and at
;;; the beginning of bodies.  Any sequence of expressions that would
;;; be allowed where the "begin" occurs is allowed.

;;; Objects with no standard print syntax, including objects containing
;;; cycles and syntax object, are allowed in quoted data as long as they
;;; are contained within a syntax form or produced by datum->syntax-object.
;;; Such objects are never copied.

;;; All identifiers that don't have macro definitions and are
;;; not bound lexically are assumed to be global variables

;;; Top-level definitions of macro-introduced identifiers are allowed.
;;; This may not be appropriate for implementations in which the
;;; model is that bindings are created by definitions, as opposed to
;;; one in which initial values are assigned by definitions.

;;; Top-level variable definitions of syntax keywords is not permitted.
;;; Any solution allowing this would be kludgey and would yield
;;; surprising results in some cases.  We can provide an undefine-syntax
;;; form.  The questions is, should define be an implicit undefine-syntax?
;;; We've decided no for now.

;;; Identifiers and syntax objects are implemented as vectors.
;;; They should be implemented in terms of some sort of
;;; (implementation-dependent) opaque structures.

;;; This code assumes that it is possible to generate globally unique
;;; symbols.  If not, new symbols can be replaced with other unique
;;; objects during expansion and the resulting identifiers can be fed
;;; directly into the compiler or replaced with locally unique symbols in
;;; a post-pass.

(let ()

(begin
(define top-level-eval-hook
  ; for top-level macro transformers and eval-when, use default
  ; system evaluator
  (lambda (x)
    (eval `(\#noexpand ,x))))

(define local-eval-hook
  ; for local macro transformers, use intepreter
  (lambda (x)
    (interpret `(\#noexpand ,x))))

(define error-hook
   (lambda (who why what)
      (error who "~a ~s" why what)))

(define-syntax new-symbol-hook
  (syntax-rules ()
    ((_ str) (string->uninterned-symbol str))))

(define put-global-definition-hook
   (lambda (symbol binding)
      (putprop symbol '*macro-transformer* binding)))

(define-syntax get-global-definition-hook
  (lambda (x)
    (syntax-case x ()
      ((_ symbol) (syntax (\#sc-get-expander symbol))))))
)

(begin
(define-syntax build-application
  (syntax-rules ()
    ((_ fun-exp arg-exps) `(,fun-exp ,@arg-exps))))

(define-syntax build-conditional
  (syntax-rules ()
    ((_ test-exp then-exp else-exp)
     `(if ,test-exp ,then-exp ,else-exp))))

(define-syntax build-lexical-reference
  (syntax-rules ()
    ((_ var) var)))

(define-syntax build-lexical-assignment
  (syntax-rules ()
    ((_ var exp) `(set! ,var ,exp))))

(define-syntax build-global-reference
  (syntax-rules ()
    ((_ var) var)))

(define-syntax build-global-assignment
  (syntax-rules ()
    ((_ var exp) `(set! ,var ,exp))))

(define-syntax build-lambda
  (syntax-rules ()
    ((_ vars exp) `(lambda ,vars ,exp))))

(define-syntax build-case-lambda
  (syntax-rules ()
    ((_ clauses) `(case-lambda ,@clauses))))

(define-syntax build-primitive
  (syntax-rules ()
    ((_ x) `(\#primitive ,x))
    ((_ n x) `(\#primitive ,n ,x))))

(define-syntax build-data
  (syntax-rules ()
    ((_ exp)
     (let ((x exp))
       (if (self-evaluating? x)
           x
           `(quote ,x))))))

(define-syntax build-sequence
  (syntax-rules ()
    ((_ exps)
     (let ((ls exps))
       (if (null? (cdr ls))
           (car ls)
           `(begin ,@ls))))))

(define build-letrec
  (lambda (vars val-exps body-exp)
    (if (null? vars)
        body-exp
        `(letrec ,(map list vars val-exps) ,body-exp))))

(define-syntax build-global-definition
  (syntax-rules ()
    ((_ var val) `(define ,var ,val))))

(define-syntax self-evaluating?
  (syntax-rules ()
    ((_ e)
     (let ((x e)) (not (or (symbol? x) (pair? x)))))))
)

(define-syntax expand-optimize-level (syntax-rules () ((_) 3)))

; syntax objects---an opaque structure should really be used here

(define-syntax syntax-object?
  (syntax-rules ()
    ((_ e)
     (let ((x e))
       (and (vector? x)
            (fx= (vector-length x) 3)
            (eq? (vector-ref x 0) 'syntax-object))))))

(define-syntax make-syntax-object
  (syntax-rules ()
    ((_ expression wrap)
     (vector 'syntax-object expression wrap))))

(define-syntax syntax-object-expression
  (syntax-rules ()
    ((_ x) (vector-ref x 1))))

(define-syntax syntax-object-wrap
  (syntax-rules ()
    ((_ x) (vector-ref x 2))))

(define-syntax arg-check
  (syntax-rules ()
    ((_ pred? e who)
     (let ((x e))
       (if (not (pred? x)) (error-hook who "invalid argument" x))))))


;;; identifier bindings

;;; <binding> ::= (macro . <procedure>)           macros
;;;               (core . <procedure>)            core forms
;;;               (begin)                         begin
;;;               (define)                        define
;;;               (define-syntax)                 define-syntax
;;;               (local-syntax . bind)           let-syntax/letrec-syntax
;;;               (eval-when)                     eval-when
;;;               (syntax . <level>)              pattern variables
;;;               (global)                        assumed global variable
;;;               (lexical)                       lexical variables
;;;               (displaced-lexical)             displaced lexicals

;;; a macro is a user-defined syntactic-form.  a core is a system-defined
;;; syntactic form.  begin, define, define-syntax, and eval-when are
;;; treated specially since they are sensitive to whether the form is
;;; at top-level and (except for eval-when) can denote valid internal
;;; definitions.

;;; a pattern variable is a variable introduced by syntax-case and can
;;; be referenced only within a syntax form.

;;; any identifier for which no top-level syntax definition or local
;;; binding of any kind has been seen is assumed to be a global
;;; variable.  global variable definitions are currently not permitted
;;; for identifiers bound as syntax.  it would be possible to treat
;;; treating global define an implicit "undefine-syntax" but this is
;;; problematic for normal evaluation and downright scary for file
;;; compiles.

;;; a lexical variable is a lambda- or letrec-bound variable.

;;; a displaced-lexical identifier is a lexical identifier removed from
;;; it's scope by the return of a syntax object containing the identifier.
;;; a displaced lexical can also appear when a letrec-syntax-bound
;;; keyword is referenced on the rhs of one of the letrec-syntax clauses.
;;; a displaced lexical should never occur with properly written macros.

(define-syntax binding-type
  (syntax-rules ()
    ((_ x) (car x))))

(define-syntax binding-value
  (syntax-rules ()
    ((_ x) (cdr x))))

;;; lexical compile-time environments

;;; <environment>              ::= (<element>*)
;;; <element>                  ::= <variable> |
;;;                                (<symbol> . <pattern-variable binding>) |
;;;                                (<symbol> . <macro binding>)
;;; <variable>                 ::= <symbol>
;;; <pattern-variable binding> ::= (syntax . level)
;;; <macro binding>            ::= (macro . <transformer>)
;;; <transformer>              ::= <procedure>
;;; <level>                    ::= <nonnegative integer>

(define-syntax null-env
  (syntax-rules ()
    ((_) '())))

(define extend-macro-env
   (lambda (vars vals r)
      (if (null? vars)
          r
          (cons (cons (car vars) (cons 'macro (car vals)))
                (extend-macro-env (cdr vars) (cdr vals) r)))))

(define-syntax extend-var-env
  (syntax-rules ()
    ((_ vars env) (append vars env))))

(define extend-syntax-env
   (lambda (vars vals r)
      (if (null? vars)
          r
          (cons (cons (car vars) (cons 'syntax (car vals)))
                (extend-syntax-env (cdr vars) (cdr vals) r)))))

;; see explanation of lexical compile-time environments above
;; name is the var-name of id

;(define lookup
;   (lambda (name id r)
;      (if (eq? name (id-sym-name id))
;          (or (get-global-definition-hook name) '(global))
;          (let search ((r r) (name name))
;             (cond
;                ((null? r) '(displaced-lexical))
;                ((eq? (car r) name) '(lexical))
;                ((pair? (car r))
;                 (if (eq? (caar r) name) (cdar r) (search (cdr r) name)))
;                (else (search (cdr r) name)))))))

;; hope this doesn't slow things down too much.
;; supports fluid-let-syntax
(define lookup
  (lambda (name id r)
    (let search ([r r] [name name])
      (cond
        [(null? r)
         (if (eq? name (id-sym-name id))
             (or (get-global-definition-hook name) '(global))
             '(displaced-lexical))]
        [(eq? (car r) name) '(lexical)]
        [(pair? (car r))
         (if (eq? (caar r) name) (cdar r) (search (cdr r) name))]
        [else (search (cdr r) name)]))))

(define global-extend
   (lambda (type sym val)
      (put-global-definition-hook sym (cons type val))))


;;; Conceptually, identifiers are always syntax objects.  Internally,
;;; however, the wrap is sometimes maintained separately, so that
;;; symbols are also considered identifiers by id?.  Externally,
;;; they are always wrapped.

(define nonsymbol-id?
  (lambda (x)
    (and (syntax-object? x) (symbol? (syntax-object-expression x)))))

(define id?
  (lambda (x)
    (or (symbol? x)
        (and (syntax-object? x) (symbol? (syntax-object-expression x))))))

(define-syntax id-sym-name
  (syntax-rules ()
    ((_ x) (if (symbol? x) x (syntax-object-expression x)))))

;;; syntax object wraps

;;;         <wrap> ::= (<wrap element>*)
;;; <wrap element> ::= <mark> |
;;;                    (<old name> <new name> <mark>*)

(define-syntax empty-wrap
  (syntax-rules ()
    ((_) '())))

(define-syntax top-wrap
  (syntax-rules ()
    ((_) '(top))))

(define-syntax top-marked?
  (syntax-rules ()
    ((_ w) (memq 'top w))))

;;; Marks must be comparable with "eq?" and distinct from pairs and
;;; the symbol top.  We no longer use integers so that marks will
;;; remain unique even across file compiles.

(define-syntax new-mark-wrap
  (syntax-rules ()
    ((_) (list (string #\m)))))

(define make-binding-wrap
   (lambda (ids new-names w)
      (if (null? ids) w (cons (make-wrap-rib ids new-names w) w))))

(define make-wrap-rib
  (lambda (ids new-names w)
    (if (null? ids)
        '()
        (cons (call-with-values
                (lambda () (id-var-name&marks (car ids) w))
                (lambda (name marks)
                  (cons name (cons (car new-names) marks))))
              (make-wrap-rib (cdr ids) (cdr new-names) w)))))

;; OW:  how do we know it's the last element of w1 that may cancel
;;      the marks in w2?

(define join-wraps
  (lambda (w1 w2)
    (cond
      ((null? w1) w2)
      ((null? w2) w1)
      ((pair? (car w2))
       ; w2 starts with a substitution
       (cons (car w1)
             (let join ((w1 (cdr w1)) (w2 w2))
               (if (null? w1)
                   w2
                   (cons (car w1) (join (cdr w1) w2))))))
      (else
       ; w2 starts with a mark; last element of w1 may cancel
       (let join ((x1 (car w1)) (w1 (cdr w1)) (x2 (car w2)) (w2 w2))
         (if (null? w1)
             (if (eq? x1 x2)
                 (cdr w2)
                 (cons x1 w2))
             (cons x1 (join (car w1) (cdr w1) x2 w2))))))))

(define same-marks?
  (lambda (x y)
    (if (null? x)
        (null? y)
        (and (not (null? y))
             (eq? (car x) (car y))
             (same-marks? (cdr x) (cdr y))))))

(define id-var-name
   (lambda (id w)
      (cond
         ((null? w)
          (if (symbol? id)
              id
              (id-var-name
                 (syntax-object-expression id)
                 (syntax-object-wrap id))))
         ((pair? (car w))
          (call-with-values
            (lambda () (id-var-name&marks id w))
            (lambda (name marks) name)))
         (else (id-var-name id (cdr w))))))

;; OW:
;; not 100% sure what's going on here,
;; could we do it more efficiently?

(define id-var-name&marks
  (lambda (id w)
    (if (null? w)
        (if (symbol? id)
            (values id '())
            (id-var-name&marks
               (syntax-object-expression id)
               (syntax-object-wrap id)))
        (call-with-values
          (lambda () (id-var-name&marks id (cdr w)))
          (lambda (name marks)
            (let ((first (car w)))
              (if (pair? first)
                  (let search ((rib first))
                    (cond
                      ((null? rib) (values name marks))
                      ((and (eq? (caar rib) name)
                            (same-marks? marks (cddar rib)))
                       (values (cadar rib) marks))
                      (else (search (cdr rib)))))
                  (values name
                          (if (or (null? marks)
                                  (not (eq? first (car marks))))
                              (cons first marks)
                              (cdr marks))))))))))

;;; free-id=? must be passed fully wrapped ids since (free-id=? x y)
;;; may be true even if (free-id=? (wrap x w) (wrap y w)) is not.

(define free-id=?
  (lambda (i j)
    (and (eq? (id-sym-name i) (id-sym-name j))
         (eq? (id-var-name i (empty-wrap)) (id-var-name j (empty-wrap))))))

;;; bound-id=? may be passed unwrapped (or partially wrapped) ids as
;;; long as the missing portion of the wrap is common to both of the ids
;;; since (bound-id=? x y) iff (bound-id=? (wrap x w) (wrap y w))

(define bound-id=?
  (lambda (i j)
    (and (eq? (id-sym-name i) (id-sym-name j))
         (call-with-values
           (lambda () (id-var-name&marks i (empty-wrap)))
           (lambda (i-name i-marks)
             (call-with-values
               (lambda () (id-var-name&marks j (empty-wrap)))
               (lambda (j-name j-marks)
                 (and (eq? i-name j-name)
                      (same-marks? i-marks j-marks)))))))))

;;; "valid-bound-ids?" returns #t if it receives a list of unique ids.
;;; valid-bound-ids? may be passed unwrapped (or partially wrapped) ids
;;; as long as the missing portion of the wrap is common to all of the
;;; ids.

(define valid-bound-ids?
  (lambda (ids)
     (and (let all-ids? ((ids ids))
            (or (null? ids)
                (and (id? (car ids))
                     (all-ids? (cdr ids)))))
          (unique-bound-ids? ids))))

;;; unique-bound-ids? expects a list of ids and returns #t if there are
;;; no duplicates.  It is quadratic on the length of the id list; long
;;; lists could be sorted to make it more efficient.  unique-bound-ids?
;;; may be passed unwrapped (or partially wrapped) ids as long as the
;;; missing portion of the wrap is common to all of the ids.

(define unique-bound-ids?
  (lambda (ids)
    (let unique? ((ids ids))
      (or (null? ids)
          (and (not (bound-id-member? (car ids) (cdr ids)))
               (unique? (cdr ids)))))))

(define bound-id-member?
   (lambda (x list)
      (and (not (null? list))
           (or (bound-id=? x (car list))
               (bound-id-member? x (cdr list))))))

;;; wrapping expressions and identifiers

(define wrap
   (lambda (x w)
      (cond
         ((null? w) x)
         ((syntax-object? x)
          (make-syntax-object (syntax-object-expression x)
                              (join-wraps w (syntax-object-wrap x))))
         ((null? x) x)
         (else (make-syntax-object x w)))))


;;; expanding

(define chi-sequence
  (lambda (body r w)
    (build-sequence
      (let dobody ((body body) (r r) (w w))
        (if (null? body)
            '()
            (let ((first (chi (car body) r w)))
              (cons first (dobody (cdr body) r w))))))))

(define chi-top-sequence
  (lambda (body r w m esew)
    (build-sequence
      (let dobody ((body body) (r r) (w w) (m m) (esew esew))
        (if (null? body)
            '()
            (let ((first (chi-top (car body) r w m esew)))
              (cons first (dobody (cdr body) r w m esew))))))))

(define chi-install-global
  (lambda (name e)
    (build-application
      (build-primitive (expand-optimize-level)
        '\#install-global-transformer)
       (list (build-data name) e))))
                            
(define chi-when-list
  (lambda (e when-list w)
    (let ((when-list (strip when-list w)))
      (if (andmap (lambda (x) (memq x '(compile eval load))) when-list)
          when-list
          (syntax-error e "invalid eval-when situation")))))

;;; syntax-type returns four values: type, value, e, and w.  The first
;;; two are described in the table below.
;;;
;;;    type                   value         explanation
;;;    -------------------------------------------------------------------
;;;    core                   procedure     core form (including singleton)
;;;    lexical                name          lexical variable reference
;;;    global                 name          global variable reference
;;;    begin                  none          begin keyword
;;;    define                 none          define keyword
;;;    define-syntax          none          define-syntax keyword
;;;    local-syntax           bind          letrec-syntax/let-syntax keyword
;;;    eval-when              none          eval-when keyword
;;;    syntax                 level         pattern variable
;;;    displaced-lexical      none          displaced lexical identifier
;;;    lexical-call           name          call to lexical variable
;;;    global-call            name          call to global variable
;;;    call                   none          any other call
;;;    begin-form             none          begin expression
;;;    define-form            id            variable definition
;;;    define-syntax-form     id            syntax definition
;;;    local-syntax-form      bind          syntax definition
;;;    eval-when-form         none          eval-when form
;;;    constant               none          self-evaluating datum
;;;    other                  none          anything else
;;;
;;; For define-form and define-syntax-form, e is the rhs expression.
;;; For all others, e is the entire form.  w is the wrap for e.
;;;
;;; syntax-type expands macros and unwraps as necessary to get to
;;; one of the forms above.  It also parses define and define-syntax
;;; forms, although perhaps this should be done by the consumer.

(define syntax-type
  (lambda (e r w)
    (cond
      ((id? e)
       (let* ((n (id-var-name e w))
              (b (lookup n e r))
              (type (binding-type b)))
         (case type
           ((lexical global) (values type n e w))
           ((macro)
            (call-with-values
              (lambda () (chi-macro (binding-value b) e r w))
              (lambda (e w) (syntax-type e r w))))
           (else (values type (binding-value b) e w)))))
      ((pair? e)
       (let ((first (car e)))
         (if (id? first)
             (let* ((n (id-var-name first w))
                    (b (lookup n first r))
                    (type (binding-type b)))
               (case type
                 ((lexical) (values 'lexical-call n e w))
                 ((global) (values 'global-call n e w))
                 ((macro)
                  (call-with-values
                    (lambda () (chi-macro (binding-value b) e r w))
                    (lambda (e w) (syntax-type e r w))))
                 ((core) (values type (binding-value b) e w))
                 ((local-syntax)
                  (values 'local-syntax-form (binding-value b) e w))
                 ((begin) (values 'begin-form #f e w))
                 ((eval-when) (values 'eval-when-form #f e w))
                 ((define)
                  (syntax-case e ()
                    ((_ name val)
                     (id? (syntax name))
                     (values 'define-form (syntax name) (syntax val) w))
                    ((_ (name . args) e1 e2 ...)
                     (and (id? (syntax name))
                          (valid-bound-ids? (lambda-var-list (syntax args))))
                     ; need #%lambda here...
                     (values 'define-form (wrap (syntax name) w)
                       (cons (syntax lambda) (wrap (syntax (args e1 e2 ...)) w))
                       (empty-wrap)))
                    ((_ name)
                     (id? (syntax name))
                     (values 'define-form (wrap (syntax name) w)
                       (syntax (#%void)) (empty-wrap)))))
                 ((define-syntax)
                  (syntax-case e ()
                    ((_ name val)
                     (id? (syntax name))
                     (values 'define-syntax-form (syntax name)
                       (syntax val) w))))
                 (else (values 'call #f e w))))
             (values 'call #f e w))))
      ((syntax-object? e)
       (syntax-type (syntax-object-expression e)
                    r
                    (join-wraps w (syntax-object-wrap e))))
      ((self-evaluating? e) (values 'constant #f e w))
      (else (values 'other #f e w)))))

(define chi-top
  (lambda (e r w m esew)
    (define-syntax eval-if-c&e
      (syntax-rules ()
        ((_ m e)
         (let ((x e))
           (if (eq? m 'c&e) (top-level-eval-hook x))
           x))))
    (call-with-values
      (lambda () (syntax-type e r w))
      (lambda (type value e w)
        (case type
          ((begin-form)
           (syntax-case e ()
             ((_) (chi-void))
             ((_ e1 e2 ...)
              (chi-top-sequence (syntax (e1 e2 ...)) r w m esew))))
          ((local-syntax-form)
           (chi-local-syntax value e r w
             (lambda (body r w)
               (chi-top-sequence body r w m esew))))
          ((eval-when-form)
           (syntax-case e ()
             ((_ (s ...) e1 e2 ...)
              (let ((when-list (chi-when-list e (syntax (s ...)) w))
                    (body (syntax (e1 e2 ...))))
                (cond
                  ((eq? m 'e)
                   (if (memq 'eval when-list)
                       (chi-top-sequence body r w 'e '(eval))
                       (chi-void)))
                  ((memq 'load when-list)
                   (if (or (memq 'compile when-list)
                           (and (eq? m 'c&e) (memq 'eval when-list)))
                       (chi-top-sequence body r w 'c&e '(compile load))
                       (if (memq m '(c c&e))
                           (chi-top-sequence body r w 'c '(load))
                           (chi-void))))
                  ((or (memq 'compile when-list)
                       (and (eq? m 'c&e) (memq 'eval when-list)))
                   (top-level-eval-hook
                     (chi-top-sequence body r w 'e '(eval))))
                  (else (chi-void)))))))
          ((define-syntax-form)
           (let ((n (id-var-name value w)))
             (case m
               ((c) 
                (if (memq 'compile esew)
                    (let ((e (chi-install-global n (chi e (null-env) w))))
                      (top-level-eval-hook e)
                      (if (memq 'load esew) e (chi-void)))
                    (if (memq 'load esew)
                        (chi-install-global n (chi e (null-env) w))
                        (chi-void))))
               ((c&e)
                (let ((e (chi-install-global n (chi e (null-env) w))))
                  (top-level-eval-hook e)
                  e))
               (else
                (if (memq 'eval esew)
                    (top-level-eval-hook
                      (chi-install-global n (chi e (null-env) w))))
                (chi-void)))))
          ((define-form)
           (let ((n (id-var-name value w)))
             (case (binding-type (lookup n value r))
               ((global)
                (eval-if-c&e m
                  (build-global-definition n
                    (chi e r w))))

;; NEW
               ((displaced-lexical)
                (syntax-error (wrap value w)
                  "identifier out of context"))

               (else (syntax-error (wrap value w)
                       "cannot define keyword at top level")))))
          (else (eval-if-c&e m (chi e r w))))))))

(define chi
  (lambda (e r w)
    (call-with-values
      (lambda () (syntax-type e r w))
      (lambda (type value e w)
        (case type
          ((lexical) (build-lexical-reference value))
          ((core) (value e r w))
          ((lexical-call)
           (chi-application (build-lexical-reference value) e r w))
          ((global-call)
           (chi-application (build-global-reference value) e r w))
          ((constant) (build-data e))
          ((global) (build-global-reference value))
          ((call) (chi-application (chi (car e) r w) e r w))
          ((begin-form)
           (syntax-case e ()
             ((_ e1 e2 ...) (chi-sequence (syntax (e1 e2 ...)) r w))))
          ((local-syntax-form)
           (chi-local-syntax value e r w chi-sequence))
          ((eval-when-form)
           (syntax-case e ()
             ((_ (s ...) e1 e2 ...)
              (let ((when-list (chi-when-list e (syntax (s ...)) w)))
                (if (memq 'eval when-list)
                    (chi-sequence (syntax (e1 e2 ...)) r w)
                    (chi-void))))))
          ((define-form define-syntax-form)
           (syntax-error value "invalid context for definition of"))
          ((syntax)
           (syntax-error (wrap e w)
             "reference to pattern variable outside syntax form"))
          ((displaced-lexical)
           (syntax-error (wrap e w)
             "reference to identifier outside its scope"))
          (else (syntax-error (wrap e w))))))))

(define chi-application
  (lambda (x e r w)
    (syntax-case e ()
      ((e0 e1 ...)
       (build-application x
         (map (lambda (e) (chi e r w)) (syntax (e1 ...))))))))

(define chi-macro
  (lambda (p e r w)
    ;; RKD: we should come up with a way to detect raw symbols in
    ;;      the output of a macro that doesn't complain about symbols
    ;;      that are ultimately quoted in the output.
    (define check-macro-output
      (lambda (x)
        (cond ((pair? x) (check-macro-output (car x))
                           (check-macro-output (cdr x)))
              ((syntax-object? x))
              ((vector? x)
               (let ((n (vector-length x)))
                 (do ((i 0 (fx+ i 1)))
                     ((fx= i n))
                     (check-macro-output (vector-ref x i)))))
              ((symbol? x)
               (syntax-error x "encountered raw symbol")))))
    (let ((mw (new-mark-wrap)))
      (let ((x (p (wrap e (join-wraps mw w)))))
        (check-macro-output x)
        (values x mw)))))

(define chi-body
  ;; We use a partial wrapper containing substitutions for the
  ;; identifiers defined so far on a left-right traversal of the
  ;; body
  ;;   (1) to allow the first nondefinition form to be a call to
  ;;       one of the defined ids even if the id previously denoted a
  ;;       definition keyword or keyword for a macro expanding into a
  ;;       definition;
  ;;   (2) to prevent subsequent definition forms (but unfortunately
  ;;       not earlier ones) and the first nondefinition form from
  ;;       confusing one of the bound identifiers for an auxiliary
  ;;       keyword; and
  ;;   (3) so that we do not need to restart the expansion of the
  ;;       first nondefinition form, which is problematic anyway
  ;;       since it might be the first element of a begin that we
  ;;       have just spliced into the body (meaning if we restarted,
  ;;       we'd really need to restart with the begin or the macro
  ;;       call that expanded into the begin, and we'd have to give
  ;;       up allowing (begin <defn>+ <expr>+), which is itself
  ;;       problematic since we don't know if a begin contains only
  ;;       definitions until we've expanded it).
  ;; Unfortunately, each form that is a macro call up through the
  ;; first nondefinition will be wrapped both with the partial wrap
  ;; (perhaps multiple times) and the final wrap.
  ;;
  ;; In processing the forms of the body, we create a new environment
  ;; containing a placeholder for the bindings we will add later and
  ;; associate this environment with each form.  In processing a
  ;; let-syntax or letrec-syntax, the associated environment may be
  ;; augmented with local keyword bindings, so the environment may
  ;; be different for different forms in the body.  Once we have
  ;; gathered up all of the definitions, we evaluate the transformer
  ;; expressions and splice into r at the placeholder the new variable
  ;; and keyword bindings.  This allows let-syntax or letrec-syntax
  ;; forms local to a portion or all of the body to shadow the
  ;; definition bindings.
  ;; 
  ;; Subforms of a begin, let-syntax, or letrec-syntax are spliced
  ;; into the body.
  (lambda (body source old-r sw)
    (let ((r (cons "placeholder" old-r)))
      (let parse ((body (map (lambda (x) (cons r (wrap x sw))) body))
                  (var-ids '()) (new-var-names '()) (var-vals '())
                  (macro-ids '()) (new-macro-names '()) (macro-vals '())
                  (pw (empty-wrap)))
        (if (null? body)
            (syntax-error (wrap source sw) "no expressions in body")
            (let ((e (cdar body)) (er (caar body)))
              (call-with-values
                (lambda () (syntax-type e er pw))
                (lambda (type value e w)
                  (case type
                    ((define-form)
                     (let ((id (wrap value w)))
                       (let ((name (gen-var id)))
                         (parse (cdr body)
                           (cons id var-ids)
                           (cons name new-var-names)
                           (cons (cons er (wrap e w)) var-vals)
                           macro-ids new-macro-names macro-vals
                           (make-binding-wrap (list id) (list name) pw)))))
                    ((define-syntax-form)
                     (let ((id (wrap value w)))
                       (let ((name (gen-var id)))
                         (parse (cdr body)
                           var-ids new-var-names var-vals
                           (cons id macro-ids)
                           (cons name new-macro-names)
                           (cons (wrap e w) macro-vals)
                           (make-binding-wrap (list id) (list name) pw)))))
                    ((begin-form)
                     (syntax-case e ()
                       ((_ e1 ...)
                        (parse (let f ((forms (syntax (e1 ...))))
                                 (if (null? forms)
                                     (cdr body)
                                     (cons (cons er (wrap (car forms) w))
                                           (f (cdr forms)))))
                          var-ids new-var-names var-vals
                          macro-ids new-macro-names macro-vals
                          pw))))
                    ((local-syntax-form)
                     (chi-local-syntax value e er w
                       (lambda (forms er w)
                         (parse (let f ((forms forms))
                                  (if (null? forms)
                                      (cdr body)
                                      (cons (cons er (wrap (car forms) w))
                                            (f (cdr forms)))))
                           var-ids new-var-names var-vals
                           macro-ids new-macro-names macro-vals
                           pw))))
                    (else
                     (if (valid-bound-ids?
                           (map (lambda (x) (wrap x pw))
                                (append macro-ids var-ids)))
                         (begin
                           (set-cdr! r
                             (extend-macro-env
                               new-macro-names
                               (map (lambda (x)
                                      (eval-local-transformer
                                        (chi x r pw)))
                                    macro-vals)
                               (extend-var-env new-var-names (cdr r))))
                           (build-letrec
                             new-var-names
                             (map (lambda (x) (chi (cdr x) (car x) pw)) var-vals)
                             (build-sequence
                                (map (lambda (x) (chi (cdr x) (car x) pw))
                                     (cons (cons er (wrap e w)) (cdr body))))))
                         (syntax-error (wrap source sw)
              "invalid or duplicate identifier in definition"))))))))))))

(define chi-lambda-clause
   (lambda (e c r w k)
      (syntax-case c ()
         (((id ...) e1 e2 ...)
          (if (not (valid-bound-ids? (syntax (id ...))))
              (syntax-error (wrap e w) "invalid parameter list in")
              (let ((new-vars (map gen-var (syntax (id ...)))))
                 (k new-vars
                    (chi-body (syntax (e1 e2 ...))
                              e
                              (extend-var-env new-vars r)
                              (make-binding-wrap (syntax (id ...))
                                                 new-vars w))))))
         ((ids e1 e2 ...)
          (let ((old-ids (lambda-var-list (syntax ids))))
             (if (not (valid-bound-ids? old-ids))
                 (syntax-error (wrap e w) "invalid parameter list in")
                 (let ((new-vars (map gen-var old-ids)))
                    (k (let f ((ls1 (cdr new-vars)) (ls2 (car new-vars)))
                          (if (null? ls1)
                              ls2
                              (f (cdr ls1) (cons (car ls1) ls2))))
                       (chi-body (syntax (e1 e2 ...))
                                 e
                                 (extend-var-env new-vars r)
                                 (make-binding-wrap old-ids new-vars w)))))))
         (_ (syntax-error (wrap e w))))))

;; hmm.  problem is that lookup assumes global anything where id-sym-name is
;; eq? to id-var-name.  so this screws up toplevel stuff...

;; hmm.  why the hell did I even bother with the map for new-vars
;; when (eq? bind 'fluid)?
;; need to do this because extend-macro-env doesn't unwrap

;; hmm. not sure if my brain is on, but why bother with the wrap when fluid?
;; (changing this seems to work)

(define chi-local-syntax
  (lambda (bind e r w k)
    (syntax-case e ()
      ((_ ((var val) ...) e1 e2 ...)
       (valid-bound-ids? (syntax (var ...)))
       (let ((new-vars (map (if (eq? bind 'fluid)
                                (lambda (x) (id-var-name x w))
                                gen-var)
                            (syntax (var ...)))))
         (let ((new-w (if (eq? bind 'fluid)
                          w
                         (make-binding-wrap (syntax (var ...)) new-vars w))))
           (k (syntax (e1 e2 ...))
              (extend-macro-env
                new-vars
                (let ((w (if (eq? bind 'rec) new-w w)))
                  (map (lambda (x)
                         (eval-local-transformer (chi x r w)))
                       (syntax (val ...))))
                r)
              new-w))))
      (_ (syntax-error (wrap e w))))))

(define eval-local-transformer
  (lambda (exp)
    (let ((p (local-eval-hook exp)))
      (if (procedure? p)
          p
          (syntax-error p "nonprocedure transfomer")))))

(define chi-void
  (lambda ()
    (build-application
      (build-primitive (expand-optimize-level) 'void)
      '())))

(define ellipsis?
  (lambda (x)
    (and (nonsymbol-id? x)
         (free-id=? x (syntax (... ...))))))

;;; data

(define strip
  (lambda (x w)
    (if (top-marked? w)
        x
        (let strip ((x x))
          (cond
            ((syntax-object? x)
             (if (top-marked? (syntax-object-wrap x))
                 (syntax-object-expression x)
                 (strip (syntax-object-expression x))))
            ((pair? x)
             (let ((a (strip (car x))) (d (strip (cdr x))))
                (if (and (eq? a (car x)) (eq? d (cdr x)))
                    x
                    (cons a d))))
            ((vector? x)
             (let ((old (vector->list x)))
                (let ((new (map strip old)))
                   (if (andmap eq? old new) x (list->vector new)))))
            (else x))))))

;;; lexical variables

(define gen-var
  (lambda (id)
    (let ((sym (id-sym-name id)))
      (let ((x (new-symbol-hook (symbol->string sym))))
        (putprop x '*real-name* sym)
        x))))

(define lambda-var-list
  (lambda (vars)
    (let lvl ((vars vars) (ls '()) (w (empty-wrap)))
       (cond
         ((pair? vars) (lvl (cdr vars) (cons (wrap (car vars) w) ls) w))
         ((id? vars) (cons (wrap vars w) ls))
         ((null? vars) ls)
         ((syntax-object? vars)
          (lvl (syntax-object-expression vars)
               ls
               (join-wraps w (syntax-object-wrap vars))))
       ; include anything else to be caught by subsequent error
       ; checking
         (else (cons vars ls))))))

;;; core transformers

(global-extend 'local-syntax 'letrec-syntax 'rec)
(global-extend 'local-syntax 'let-syntax #f)
(global-extend 'local-syntax 'fluid-let-syntax 'fluid)

(global-extend 'core 'quote
   (lambda (e r w)
      (syntax-case e ()
         ((_ e) (build-data (strip (syntax e) w)))
         (_ (syntax-error (wrap e w))))))

(global-extend 'core 'syntax
  (let ()
    (define gen-syntax
      (lambda (src e r maps ellipsis?)
        (if (id? e)
            (let ((var (id-var-name e (empty-wrap))))
              (let ((b (lookup var e r)))
                (if (eq? (binding-type b) 'syntax)
                    (call-with-values
                      (lambda () (gen-ref src var (binding-value b) maps))
                      (lambda (var maps) (values `(ref ,var) maps)))
                    (if (ellipsis? e)
                        (syntax-error src "misplaced ellipsis in syntax form")
                        (values `(quote ,e) maps)))))
            (syntax-case e ()
              ((dots e)
               (ellipsis? (syntax dots))
               (gen-syntax src (syntax e) r maps (lambda (x) #f)))
              ((x dots . y)
               ; this could be about a dozen lines of code, except that we
               ; choose to handle (syntax (x ... ...)) forms
               (ellipsis? (syntax dots))
               (let f ((y (syntax y))
                       (k (lambda (maps)
                            (call-with-values
                              (lambda ()
                                (gen-syntax src (syntax x) r
                                  (cons '() maps) ellipsis?))
                              (lambda (x maps)
                                (if (null? (car maps))
                                    (syntax-error src
                                      "extra ellipsis in syntax form")
                                    (values (gen-map x (car maps))
                                            (cdr maps))))))))
                 (syntax-case y ()
                   ((dots . y)
                    (ellipsis? (syntax dots))
                    (f (syntax y)
                       (lambda (maps)
                         (call-with-values
                           (lambda () (k (cons '() maps)))
                           (lambda (x maps)
                             (if (null? (car maps))
                                 (syntax-error src
                                   "extra ellipsis in syntax form")
                                 (values (gen-mappend x (car maps))
                                         (cdr maps))))))))
                   (_ (call-with-values
                        (lambda () (gen-syntax src y r maps ellipsis?))
                        (lambda (y maps)
                          (call-with-values
                            (lambda () (k maps))
                            (lambda (x maps)
                              (values (gen-append x y) maps)))))))))
              ((x . y)
               (call-with-values
                 (lambda () (gen-syntax src (syntax x) r maps ellipsis?))
                 (lambda (x maps)
                   (call-with-values
                     (lambda () (gen-syntax src (syntax y) r maps ellipsis?))
                     (lambda (y maps) (values (gen-cons x y) maps))))))
              (#(e1 e2 ...)
               (call-with-values
                 (lambda ()
                   (gen-syntax src (syntax (e1 e2 ...)) r maps ellipsis?))
                 (lambda (e maps) (values (gen-vector e) maps))))
              (_ (values `(quote ,e) maps))))))

    (define gen-ref
      (lambda (src var level maps)
        (if (fx= level 0)
            (values var maps)
            (if (null? maps)
                (syntax-error src "missing ellipsis in syntax form")
                (call-with-values
                  (lambda () (gen-ref src var (fx- level 1) (cdr maps)))
                  (lambda (outer-var outer-maps)
                    (let ((b (assq outer-var (car maps))))
                      (if b
                          (values (cdr b) maps)
                          (let ((inner-var
                                 (new-symbol-hook (symbol->string var))))
                            (values inner-var
                                    (cons (cons (cons outer-var inner-var)
                                                (car maps))
                                          outer-maps)))))))))))

    (define gen-mappend
      (lambda (e map-env)
        `(apply (primitive append) ,(gen-map e map-env))))

    (define gen-map
      (lambda (e map-env)
        (let ((formals (map cdr map-env))
              (actuals (map (lambda (x) `(ref ,(car x))) map-env)))
          (cond
            ((eq? (car e) 'ref)
             ; identity map equivalence:
             ; (map (lambda (x) x) y) == y
             (car actuals))
            ((andmap
                (lambda (x) (and (eq? (car x) 'ref) (memq (cadr x) formals)))
                (cdr e))
             ; eta map equivalence:
             ; (map (lambda (x ...) (f x ...)) y ...) == (map f y ...)
             `(map (primitive ,(car e))
                   ,@(map (let ((r (map cons formals actuals)))
                            (lambda (x) (cdr (assq (cadr x) r))))
                          (cdr e))))
            (else `(map (lambda ,formals ,e) ,@actuals))))))

    (define gen-cons
      (lambda (x y)
        (cond
          ((eq? (car y) 'quote)
           (if (eq? (car x) 'quote)
               `(quote (,(cadr x) . ,(cadr y)))
               (if (eq? (cadr y) '())
                   `(list ,x)
                   `(list* ,x ,y))))
          ((memq (car y) '(list list*))
           `(,(car y) ,x ,@(cdr y)))
          (else `(list* ,x ,y)))))

    (define gen-append
      (lambda (x y)
        (if (equal? y '(quote ()))
            x
            `(append ,x ,y))))

    (define gen-vector
      (lambda (x)
        (cond
          ((eq? (car x) 'list) `(vector ,@(cdr x)))
          ((eq? (car x) 'quote) `(quote #(,@(cadr x))))
          (else `(list->vector ,x)))))

    (define regen
      (lambda (x)
        (case (car x)
          ((ref) (build-lexical-reference (cadr x)))
          ((primitive) (build-primitive (expand-optimize-level) (cadr x)))
          ((quote) (build-data (cadr x)))
          ((lambda) (build-lambda (cadr x) (regen (caddr x))))
          ((map) (let ((ls (map regen (cdr x))))
                   (build-application
                     (if (fx= (length ls) 2)
                         (build-primitive (expand-optimize-level) 'map)
                         (build-primitive 2 'map)) ; require error check
                     ls)))
          (else (build-application
                  (build-primitive (expand-optimize-level) (car x))
                  (map regen (cdr x)))))))

    (lambda (e r w)
      (let ((e (wrap e w)))
        (syntax-case e ()
          ((_ x)
           (call-with-values
             (lambda () (gen-syntax e (syntax x) r '() ellipsis?))
             (lambda (e maps) (regen e))))
          (_ (syntax-error e)))))))

(global-extend 'core '\#primitive
  ; should verify that symbol names a valid primitive
  (lambda (e r w)
    (syntax-case e ()
      ((_ x)
       (id? (syntax x))
       (build-primitive (strip (syntax x) w)))
      ((_ n x)
       (and (memv (syntax-object->datum (syntax n)) '(2 3))
            (id? (syntax x)))
       (build-primitive (strip (syntax n) w) (strip (syntax x) w)))
      (_ (syntax-error (wrap e w))))))

(global-extend 'core '\#noexpand
  (lambda (e r w)
    (syntax-case e ()
      ((_ x) (strip (syntax x) w))
      (_ (syntax-error (wrap e w))))))

(global-extend 'core 'lambda
   (lambda (e r w)
      (syntax-case e ()
         ((_ . c)
          (chi-lambda-clause e (syntax c) r w
            (lambda (vars body) (build-lambda vars body)))))))

(global-extend 'core 'case-lambda
   (lambda (e r w)
      (syntax-case e ()
         ((_ c ...)
          (build-case-lambda
             (map (lambda (c) (chi-lambda-clause e c r w list))
                  (syntax (c ...))))))))

(global-extend 'core 'letrec
   (lambda (e r w)
      (syntax-case e ()
         ((_ ((id val) ...) e1 e2 ...)
          (valid-bound-ids? (syntax (id ...)))
          (let ((new-vars (map gen-var (syntax (id ...)))))
             (let ((w (make-binding-wrap (syntax (id ...)) new-vars w))
                   (r (extend-var-env new-vars r)))
                (build-letrec
                   new-vars
                   (map (lambda (x) (chi x r w)) (syntax (val ...)))
                   (chi-body (syntax (e1 e2 ...)) e r w)))))
         (_ (syntax-error (wrap e w))))))

(global-extend 'core 'if
   (lambda (e r w)
      (syntax-case e ()
         ((_ test then)
          (build-conditional
             (chi (syntax test) r w)
             (chi (syntax then) r w)
             (chi-void)))
         ((_ test then else)
          (build-conditional
             (chi (syntax test) r w)
             (chi (syntax then) r w)
             (chi (syntax else) r w)))
         (_ (syntax-error (wrap e w))))))

(global-extend 'core 'foreign-procedure
  (lambda (e r w)
    (syntax-case e ()
      ((_ name (arg ...) result)
       `(foreign-procedure
          ,(chi (syntax name) r w)
          ,(strip (syntax (arg ...)) w)
          ,(strip (syntax result) w))))))

(global-extend 'core 'set!
   (lambda (e r w)
      (syntax-case e ()
         ((_ id val)
          (id? (syntax id))
          (let ((val (chi (syntax val) r w))
                (n (id-var-name (syntax id) w)))
             (case (binding-type (lookup n (syntax id) r))
                ((lexical) (build-lexical-assignment n val))
                ((global) (build-global-assignment n val))
                (else (syntax-error (wrap e w))))))
         (_ (syntax-error (wrap e w))))))

(global-extend 'begin 'begin '())

(global-extend 'define 'define '())

(global-extend 'define-syntax 'define-syntax '())

(global-extend 'eval-when 'eval-when '())

(global-extend 'core 'syntax-case
  (let ()
    (define convert-pattern
      ; accepts pattern & keys
      ; returns syntax-dispatch pattern & ids
      (lambda (pattern keys)
        (let cvt ((p pattern) (n 0) (ids '()))
          (if (id? p)
              (if (bound-id-member? p keys)
                  (values (vector 'free-id p) ids)
                  (values 'any (cons (cons p n) ids)))
              (syntax-case p ()
                ((x dots)
                 (ellipsis? (syntax dots))
                 (call-with-values
                   (lambda () (cvt (syntax x) (fx+ n 1) ids))
                   (lambda (p ids)
                     (values (if (eq? p 'any) 'each-any (vector 'each p))
                             ids))))
                ((x . y)
                 (call-with-values
                   (lambda () (cvt (syntax y) n ids))
                   (lambda (y ids)
                     (call-with-values
                       (lambda () (cvt (syntax x) n ids))
                       (lambda (x ids)
                         (values (cons x y) ids))))))
                (() (values '() ids))
                (#(x ...)
                 (call-with-values
                   (lambda () (cvt (syntax (x ...)) n ids))
                   (lambda (p ids) (values (vector 'vector p) ids))))
                (x (values (vector 'atom (strip p (empty-wrap))) ids)))))))

    (define build-dispatch-call
      (lambda (pvars exp y r)
        (let ((vars (map car pvars)) (vals (map cdr pvars)))
          (let ((new-vars (map gen-var vars)))
            (build-application (build-primitive (expand-optimize-level) 'apply)
              (list (build-lambda new-vars
                      (chi exp
                           (extend-syntax-env new-vars vals r)
                           (make-binding-wrap vars new-vars (empty-wrap))))
                    y))))))

    (define gen-clause
      (lambda (x keys clauses r pat fender exp)
        (call-with-values
          (lambda () (convert-pattern pat keys))
          (lambda (p pvars)
            (cond
              ((not (unique-bound-ids? (map car pvars)))
               (syntax-error pat
                 "duplicate pattern variable in syntax-case pattern"))
              ((not (andmap (lambda (x) (not (ellipsis? (car x)))) pvars))
               (syntax-error pat
                 "misplaced ellipsis in syntax-case pattern"))
              (else
               (let ((y (new-symbol-hook "y")))
                 ; fat finger binding and references to temp variable y
                 (build-application
                   (build-lambda (list y)
                     (let ((y (build-lexical-reference y)))
                       (build-conditional
                         (syntax-case fender ()
                           (#t y)
                           (_ (build-conditional
                                y
                                (build-dispatch-call pvars fender y r)
                                (build-data #f))))
                         (build-dispatch-call pvars exp y r)
                         (gen-syntax-case x keys clauses r))))
                   (list (build-application
                           (build-primitive (expand-optimize-level)
                             '\#syntax-dispatch)
                           (list x (build-data p))))))))))))

    (define gen-syntax-case
      (lambda (x keys clauses r)
        (if (null? clauses)
            (build-application
              (build-primitive (expand-optimize-level) 'syntax-error)
              (list x))
            (syntax-case (car clauses) ()
              ((pat exp)
               (if (and (id? (syntax pat))
                        (andmap (lambda (x) (not (free-id=? (syntax pat) x)))
                          (cons (syntax (... ...)) keys)))
                   (let ((new-vars (list (gen-var (syntax pat)))))
                     (build-application
                       (build-lambda new-vars
                         (chi (syntax exp)
                              (extend-syntax-env new-vars '(0) r)
                              (make-binding-wrap (syntax (pat))
                                new-vars (empty-wrap))))
                       (list x)))
                   (gen-clause x keys (cdr clauses) r
                     (syntax pat) (syntax #t) (syntax exp))))
              ((pat fender exp)
               (gen-clause x keys (cdr clauses) r
                 (syntax pat) (syntax fender) (syntax exp)))
              (_ (syntax-error (car clauses) "invalid syntax-case clause"))))))

    (lambda (e r w)
      (let ((e (wrap e w)))
        (syntax-case e ()
          ((_ val (key ...) m ...)
           (if (andmap (lambda (x)
                         (and (id? x)
                              (not (free-id=? x (syntax (... ...))))))
                       (syntax (key ...)))
               (let ((x (new-symbol-hook "x")))
                 ; fat finger binding and references to temp variable x
                 (build-application
                   (build-lambda (list x)
                     (gen-syntax-case (build-lexical-reference x)
                       (syntax (key ...)) (syntax (m ...))
                       r))
                   (list (chi (syntax val) r (empty-wrap)))))
               (syntax-error e "invalid literals list in"))))))))

;;; exports

(set! sc-expand
  (lambda (x)
    (chi-top x (null-env) (top-wrap)
      (if \#compiling-a-file 'c 'e)
      (eval-syntax-expanders-when))))

(set! identifier?
  (lambda (x)
    (nonsymbol-id? x)))

(set! datum->syntax-object
  (lambda (id datum)
    (arg-check nonsymbol-id? id 'datum->syntax-object)
    (wrap datum (syntax-object-wrap id))))

(set! syntax-object->datum
  ; accepts any object, since syntax objects may consist partially
  ; or entirely of unwrapped, nonsymbolic data
  (lambda (x)
    (strip x (empty-wrap))))

(set! generate-temporaries
  (lambda (ls)
    (arg-check list? ls 'generate-temporaries)
    (let f ((ls ls) (i 1))
      (if (null? ls)
          '()
          (cons (wrap (new-symbol-hook (string-append "t" (number->string i)))
                      (top-wrap))
                (f (cdr ls) (fx+ i 1)))))))

(set! free-identifier=?
   (lambda (x y)
      (arg-check nonsymbol-id? x 'free-identifier=?)
      (arg-check nonsymbol-id? y 'free-identifier=?)
      (free-id=? x y)))

(set! bound-identifier=?
   (lambda (x y)
      (arg-check nonsymbol-id? x 'bound-identifier=?)
      (arg-check nonsymbol-id? y 'bound-identifier=?)
      (bound-id=? x y)))

;; RKD: we should try to get to the point where syntax errors result
;;      in warnings w/run-time error messages---this would allow us
;;      to get multiple syntax errors in a single pass.  It would
;;      be worth writing up in our syntax-case implementation article.
(set! syntax-error
   (lambda (object . messages)
      (for-each (lambda (x) (arg-check string? x 'syntax-error)) messages)
      (let ((message (if (null? messages)
                         "invalid syntax"
                         (apply string-append messages))))
         (error-hook #f message (strip object (empty-wrap))))))

(set! \#install-global-transformer
  (lambda (sym v)
    (arg-check symbol? sym '\#install-global-transformer)
    (arg-check procedure? v '\#install-global-transformer)
    (global-extend 'macro sym v)))

;;; syntax-dispatch expects an expression and a pattern.  If the expression
;;; matches the pattern a list of the matching expressions for each
;;; "any" is returned.  Otherwise, #f is returned.  (This use of #f will
;;; not work on r4rs implementations that violate the ieee requirement
;;; that #f and () be distinct.)

;;; The expression is matched with the pattern as follows:

;;; pattern:                           matches:
;;;   ()                                 empty list
;;;   any                                anything
;;;   (<pattern>1 . <pattern>2)          (<pattern>1 . <pattern>2)
;;;   each-any                           (any*)
;;;   #(free-id <key>)                   <key> with free-identifier=?
;;;   #(each <pattern>)                  (<pattern>*)
;;;   #(vector <pattern>)                (list->vector <pattern>)
;;;   #(atom <object>)                   <object> with "equal?"

;;; Vector cops out to pair under assumption that vectors are rare.  If
;;; not, should convert to:
;;;   #(vector <pattern>*)               #(<pattern>*)

(let ()

(define match-each
  (lambda (e p w)
    (cond
      ((pair? e)
       (let ((first (match (car e) p w '())))
         (and first
              (let ((rest (match-each (cdr e) p w)))
                 (and rest (cons first rest))))))
      ((null? e) '())
      ((syntax-object? e)
       (match-each (syntax-object-expression e)
                   p
                   (join-wraps w (syntax-object-wrap e))))
      (else #f))))

(define match-each-any
  (lambda (e w)
    (cond
      ((pair? e)
       (let ((l (match-each-any (cdr e) w)))
         (and l (cons (wrap (car e) w) l))))
      ((null? e) '())
      ((syntax-object? e)
       (match-each-any (syntax-object-expression e)
                       (join-wraps w (syntax-object-wrap e))))
      (else #f))))

(define match-empty
  (lambda (p r)
    (cond
      ((null? p) r)
      ((eq? p 'any) (cons '() r))
      ((pair? p) (match-empty (car p) (match-empty (cdr p) r)))
      ((eq? p 'each-any) (cons '() r))
      (else
       (case (vector-ref p 0)
         ((each) (match-empty (vector-ref p 1) r))
         ((free-id atom) r)
         ((vector) (match-empty (vector-ref p 1) r)))))))

(define match*
  (lambda (e p w r)
    (cond
      ((pair? p)
       (and (pair? e) (match (car e) (car p) w (match (cdr e) (cdr p) w r))))
      ((eq? p 'each-any)
       (let ((l (match-each-any e w))) (and l (cons l r))))
      (else
       (case (vector-ref p 0)
         ((each)
          (if (null? e)
              (match-empty (vector-ref p 1) r)
              (let ((l (match-each e (vector-ref p 1) w)))
                (and l
                     (let collect ((l l))
                       (if (null? (car l))
                           r
                           (cons (map car l) (collect (map cdr l)))))))))
         ((free-id) (and (id? e) (free-id=? (wrap e w) (vector-ref p 1)) r))
         ((atom) (and (equal? (vector-ref p 1) (strip e w)) r))
         ((vector)
          (and (vector? e) (match (vector->list e) (vector-ref p 1) w r))))))))

(define match
  (lambda (e p w r)
    (cond
      ((not r) #f)
      ((null? p) (and (null? e) r))
      ((eq? p 'any) (cons (wrap e w) r))
      ((syntax-object? e)
       (match* (syntax-object-expression e)
               p
               (join-wraps w (syntax-object-wrap e))
               r))
      (else (match* e p w r)))))

(set! \#syntax-dispatch
   (lambda (expression pattern)
      (match expression pattern (empty-wrap) '())))
)
)

(define-syntax with-syntax
   (lambda (x)
      (syntax-case x ()
         ((_ () e1 e2 ...)
          (syntax (begin e1 e2 ...)))
         ((_ ((out in)) e1 e2 ...)
          (syntax (syntax-case in () (out (begin e1 e2 ...)))))
         ((_ ((out in) ...) e1 e2 ...)
          (syntax (syntax-case (#%list in ...) ()
                     ((out ...) (begin e1 e2 ...))))))))

(define-syntax syntax-rules
  (lambda (x)
    (syntax-case x ()
      ((_ (k ...) ((keyword . pattern) template) ...)
       (syntax (lambda (x)
                (syntax-case x (k ...)
                  ((dummy . pattern) (syntax template))
                  ...)))))))

(define-syntax or
   (lambda (x)
      (syntax-case x ()
         ((_) (syntax #f))
         ((_ e) (syntax e))
         ((_ e1 e2 e3 ...)
          (syntax (let ((t e1)) (if t t (or e2 e3 ...))))))))

(define-syntax and
   (lambda (x)
      (syntax-case x ()
         ((_ e1 e2 e3 ...) (syntax (if e1 (and e2 e3 ...) #f)))
         ((_ e) (syntax e))
         ((_) (syntax #t)))))

(define-syntax let
   (lambda (x)
      (syntax-case x ()
         ((_ ((x v) ...) e1 e2 ...)
          (andmap identifier? (syntax (x ...)))
          (syntax ((lambda (x ...) e1 e2 ...) v ...)))
         ((_ f ((x v) ...) e1 e2 ...)
          (andmap identifier? (syntax (f x ...)))
          (syntax ((letrec ((f (lambda (x ...) e1 e2 ...))) f)
                    v ...))))))

(define-syntax let*
  (lambda (x)
    (syntax-case x ()
      ((let* ((x v) ...) e1 e2 ...)
       (andmap identifier? (syntax (x ...)))
       (let f ((bindings (syntax ((x v)  ...))))
         (if (null? bindings)
             (syntax (let () e1 e2 ...))
             (with-syntax ((body (f (cdr bindings)))
                           (binding (car bindings)))
               (syntax (let (binding) body)))))))))

(define-syntax cond
  (lambda (x)
    (syntax-case x ()
      ((_ m1 m2 ...)
       (let f ((clause (syntax m1)) (clauses (syntax (m2 ...))))
         (if (null? clauses)
             (syntax-case clause (else =>)
               ((else e1 e2 ...) (syntax (begin e1 e2 ...)))
               ((e0) (syntax (let ((t e0)) (if t t))))
               ((e0 => e1) (syntax (let ((t e0)) (if t (e1 t)))))
               ((e0 e1 e2 ...) (syntax (if e0 (begin e1 e2 ...))))
               (_ (syntax-error x)))
             (with-syntax ((rest (f (car clauses) (cdr clauses))))
               (syntax-case clause (else =>)
                 ((e0) (syntax (let ((t e0)) (if t t rest))))
                 ((e0 => e1) (syntax (let ((t e0)) (if t (e1 t) rest))))
                 ((e0 e1 e2 ...) (syntax (if e0 (begin e1 e2 ...) rest)))
                 (_ (syntax-error x))))))))))

; see Chez Scheme version later
;(define-syntax case
;  (lambda (x)
;    (syntax-case x ()
;      ((_ e m1 m2 ...)
;       (with-syntax
;         ((body (let f ((clause (syntax m1)) (clauses (syntax (m2 ...))))
;                  (if (null? clauses)
;                      (syntax-case clause (else)
;                        ((else e1 e2 ...) (syntax (begin e1 e2 ...)))
;                        (((k ...) e1 e2 ...)
;                         (syntax (if (memv t '(k ...)) (begin e1 e2 ...))))
;                        (_ (syntax-error x)))
;                      (with-syntax ((rest (f (car clauses) (cdr clauses))))
;                        (syntax-case clause (else)
;                          (((k ...) e1 e2 ...)
;                           (syntax (if (memv t '(k ...))
;                                       (begin e1 e2 ...)
;                                       rest)))
;                          (_ (syntax-error x))))))))
;         (syntax (let ((t e)) body)))))))
                
(define-syntax do
   (lambda (orig-x)
      (syntax-case orig-x ()
         ((_ ((var init . step) ...) (e0 e1 ...) c ...)
          (with-syntax (((step ...)
                         (map (lambda (v s)
                                 (syntax-case s ()
                                    (() v)
                                    ((e) (syntax e))
                                    (_ (syntax-error orig-x))))
                              (syntax (var ...))
                              (syntax (step ...)))))
             (syntax-case (syntax (e1 ...)) ()
                (() (syntax (let doloop ((var init) ...)
                               (if (#%not e0)
                                   (begin c ... (doloop step ...))))))
                ((e1 e2 ...)
                 (syntax (let doloop ((var init) ...)
                            (if e0
                                (begin e1 e2 ...)
                                (begin c ... (doloop step ...))))))))))))

(define-syntax quasiquote
   (letrec
      ((quasicons
        (lambda (x y)
          (with-syntax ((x x) (y y))
            (syntax-case (syntax y) (quote \#primitive list list*)
              ((quote dy)
               (syntax-case (syntax x) (quote \#primitive)
                 ((quote dx) (syntax (quote (dx . dy))))
                 (_ (if (null? (syntax dy))
                        (syntax (#%list x))
                        (syntax (#%list* x y))))))
              ((#%list . stuff) (syntax (#%list x . stuff)))
              ((#%list* . stuff) (syntax (#%list* x . stuff)))
              (else (syntax (#%list* x y)))))))
       (quasiappend
        (lambda (x y)
          (with-syntax ((x x) (y y))
            (syntax-case (syntax y) (quote \#primitive)
              ((quote ()) (syntax x))
              (_ (syntax (#%append x y)))))))
       (quasivector
        (lambda (x)
          (with-syntax ((x x))
            (syntax-case (syntax x) (quote \#primitive list)
              ((quote (x ...)) (syntax (quote #(x ...))))
              ((#%list x ...) (syntax (#%vector x ...)))
              (_ (syntax (#%list->vector x)))))))
       (quasi
        (lambda (p lev)
           (syntax-case p (unquote unquote-splicing quasiquote)
              ((unquote p)
               (if (fx= lev 0)
                   (syntax p)
                   (quasicons (syntax (quote unquote))
                              (quasi (syntax (p)) (fx- lev 1)))))
              (((unquote-splicing p) . q)
               (if (fx= lev 0)
                   (quasiappend (syntax p) (quasi (syntax q) lev))
                   (quasicons (quasicons (syntax (quote unquote-splicing))
                                         (quasi (syntax (p)) (fx- lev 1)))
                              (quasi (syntax q) lev))))
              ((quasiquote p)
               (quasicons (syntax (quote quasiquote))
                          (quasi (syntax (p)) (fx+ lev 1))))
              ((p . q)
               (quasicons (quasi (syntax p) lev) (quasi (syntax q) lev)))
              (#(x ...) (quasivector (quasi (syntax (x ...)) lev)))
              (p (syntax (quote p)))))))
    (lambda (x)
       (syntax-case x ()
          ((_ e) (quasi (syntax e) 0))))))

;;; simple delay and w/sample force & make-promise code
;
;(define-syntax delay
;   (lambda (x)
;      (syntax-case x ()
;         ((delay exp)
;          (syntax (make-promise (lambda () exp)))))))
;
;(define make-promise
;  (lambda (thunk)
;    (define value)
;    (define set? #f)
;    (lambda ()
;      (if (not set?)
;          (call-with-values thunk
;            (lambda ls
;              (if (not set?)
;                  (begin (set! value ls)
;                         (set! set? #t))))))
;      (apply values value))))
;
;(define force
;   (lambda (promise)
;      (promise)))

(define-syntax include
  (lambda (x)
    (define read-file
      (lambda (fn k)
        (let ((p (open-input-file fn)))
          (let f ((x (read p)))
            (if (eof-object? x)
                (begin (close-input-port p) '())
                (cons (datum->syntax-object k x)
                      (f (read p))))))))
    (syntax-case x ()
      ((k filename)
       (let ((fn (syntax-object->datum (syntax filename))))
         (with-syntax (((exp ...) (read-file fn (syntax k))))
           (syntax (begin exp ...))))))))

(define-syntax critical-section
   (syntax-rules ()
      [(_ e1 e2 ...)
       (dynamic-wind
          #%disable-interrupts
          (lambda () e1 e2 ...)
          #%enable-interrupts)]))

(define-syntax define-structure
   (lambda (x)
      (define construct-name
         (lambda (template-identifier . args)
            (datum->syntax-object
               template-identifier
               (string->symbol
                  (apply string-append
                         (map (lambda (x)
                                 (if (string? x)
                                     x
                                     (symbol->string (syntax-object->datum x))))
                              args))))))
      (syntax-case x ()
         ((_ (name id1 ...))
          (syntax (define-structure (name id1 ...) ())))
         ((_ (name id1 ...) ((id2 init) ...))
          (with-syntax
             ((constructor (construct-name (syntax name) "make-" (syntax name)))
              (predicate (construct-name (syntax name) (syntax name) "?"))
              ((access ...)
               (map (lambda (x) (construct-name x (syntax name) "-" x))
                    (syntax (id1 ... id2 ...))))
              ((assign ...)
               (map (lambda (x)
                       (construct-name x "set-" (syntax name) "-" x "!"))
                    (syntax (id1 ... id2 ...))))
              (structure-length
               (fx+ (length (syntax (id1 ... id2 ...))) 1))
              ((index ...)
               (let f ((i 1) (ids (syntax (id1 ... id2 ...))))
                  (if (null? ids)
                      '()
                      (cons i (f (fx+ i 1) (cdr ids)))))))
             (syntax (begin
                        (define constructor
                           (lambda (id1 ...)
                              (let* ((id2 init) ...)
                                 (#%vector 'name id1 ... id2 ...))))
                        (define predicate
                           (lambda (x)
                              (and (#%vector? x)
                                   (#3%fx= (vector-length x) structure-length)
                                   (#%eq? (vector-ref x 0) 'name))))
                        (define access
                           (lambda (x)
                              (#%vector-ref x index)))
                        ...
                        ; define macro accessors this way:
                        ; (define-syntax access
                        ;       (syntax-case x ()
                        ;          ((_ x)
                        ;           (syntax (vector-ref x index)))))
                        ; ...
                        (define assign
                           (lambda (x update)
                              (#%vector-set! x index update)))
                        ...)))))))

(define-syntax define-syntax-expander
   (lambda (x)
      (error 'define-syntax-expander
             "not supported by syntax-case expander")))

(define-syntax extend-syntax
   (lambda (x)
      (error 'extend-syntax
             "not supported by syntax-case expander")))

(define-syntax extend-syntax/code
   (lambda (x)
      (error 'extend-syntax/code
             "not supported by syntax-case expander")))

(define-syntax fluid-let
  (lambda (x)
    (syntax-case x ()
      [(_ () e1 e2 ...) (syntax (begin e1 e2 ...))]
      [(_ ([x v] ...) e1 e2 ...)
       (andmap identifier? (syntax (x ...)))
       (with-syntax ([(y ...) (generate-temporaries (syntax (x ...)))])
         (syntax
           (let ([y v] ...)
             (let ([swap (lambda ()
                            (let ([t x])
                                (set! x y)
                                (set! y t))
                              ...)])
                   (#%dynamic-wind swap (lambda () e1 e2 ...) swap)))))])))

(define-syntax parameterize
   (lambda (x)
      (syntax-case x ()
         [(_ () e1 e2 ...) (syntax (begin e1 e2 ...))]
         [(_ ([x v] ...) e1 e2 ...)
          (andmap identifier? (syntax (x ...)))
          (with-syntax ([(p ...) (generate-temporaries (syntax (x ...)))]
                        [(y ...) (generate-temporaries (syntax (x ...)))])
             (syntax
                (let ([p x] ... [y v] ...)
                   (let ([swap (lambda ()
                                  (let ([t (p)]) (p y) (set! y t)) ...)])
                      (#%dynamic-wind swap (lambda () e1 e2 ...) swap)))))])))

(define-syntax rec
  (lambda (x)
    (syntax-case x ()
      [(_ x v)
       (identifier? (syntax x))
       (syntax (letrec ([x v]) x))])))
      
(define-syntax record-case
  (let ()
    (define build-body
      (lambda (p body)
        (syntax-case p ()
          ((id . p) 
           (with-syntax ((body (build-body (syntax p) body)))
             (syntax (let ((rec (#%cdr rec)))
                       (let ((id (#%car rec)))
                         body)))))
          (() (with-syntax ((body body))
                (syntax (begin . body))))
          (id
           (with-syntax ((body body))
             (syntax (let ((id (#%cdr rec))) . body)))))))

    (define-syntax build-clause
      (lambda (x)
        (syntax-case x ()
          ((_ tag keys idspec body rest)
           (syntax
             (with-syntax ((body (build-body (syntax idspec) (syntax body))))
               (syntax (if (#%memv tag 'keys)
                           body
                           . rest))))))))
    (lambda (x)
      (syntax-case x ()
        ((_ e m1 m2 ...)
         (with-syntax
           ((body (let f ((clause (syntax m1)) (clauses (syntax (m2 ...))))
                    (if (null? clauses)
                        (syntax-case clause (else)
                          ((else e1 e2 ...) (syntax (begin e1 e2 ...)))
                          (((key ...) idspec e1 e2 ...)
                           (build-clause tag (key ...) idspec (e1 e2 ...) ()))
                          ((key idspec e1 e2 ...)
                           (build-clause tag (key) idspec (e1 e2 ...) ()))
                          (_ (syntax-error x)))
                        (with-syntax ((rest (f (car clauses) (cdr clauses))))
                          (syntax-case clause (else)
                            (((key ...) idspec e1 e2 ...)
                             (build-clause tag (key ...) idspec (e1 e2 ...)
                               (rest)))
                            ((key idspec e1 e2 ...)
                             (build-clause tag (key) idspec (e1 e2 ...) (rest)))
                            (_ (syntax-error x))))))))
           (syntax (let ((rec e))
                     (let ((tag (#%car rec)))
                       body)))))))))

(define-syntax time
   (syntax-rules ()
      [(_ e)
       (#%\#as-time-goes-by 'e (lambda () e))]))

(define-syntax trace
  (lambda (x)
   (syntax-case x ()
      [(_ x ...)
       (andmap identifier? (syntax (x ...)))
       (syntax (#%\#trace 'x ...))])))

(define-syntax trace-define
  (lambda (x)
    (syntax-case x ()
      [(_ name val)
       (identifier? (syntax name))
       (syntax (define name (#%\#trace-closure 'name val)))]
      [(_ (name . idspec) e1 e2 ...)
       (identifier? (syntax name))
       (syntax (define name (trace-lambda name idspec e1 e2 ...)))])))

(define-syntax trace-define-syntax
  (lambda (x)
    (syntax-case x ()
      [(_ name val)
       (identifier? (syntax name))
       (syntax (define-syntax name (#%\#trace-closure 'name val)))])))

(define-syntax trace-lambda
   (syntax-rules ()
      [(trace-lambda name idspec e1 e2 ...)
       (#%\#trace-closure 'name (lambda idspec e1 e2 ...))]))

(define-syntax trace-let
   (syntax-rules ()
      [(trace-let name ((x v) ...) e1 e2 ...)
       ((rec name (trace-lambda name (x ...) e1 e2 ...)) v ...)]))

(define-syntax unless
   (syntax-rules ()
      [(_ e0 e1 e2 ...)
       (if (#%not e0) (begin e1 e2 ...))]))

(define-syntax unquote
   (lambda (x) 
      (syntax-case x ()
         [(_ e)
          (error 'unquote
                 "expression ,~s not valid outside of quasiquote"
                 (syntax-object->datum (syntax e)))])))

(define-syntax unquote-splicing
   (lambda (x) 
      (syntax-case x ()
         [(_ e)
          (error 'unquote-splicing
                 "expression ,@~s not valid outside of quasiquote"
                 (syntax-object->datum (syntax e)))])))

(define-syntax untrace
  (lambda (x)
    (syntax-case x ()
      [(untrace x ...)
       (andmap symbol? '(x ...))
       (syntax (#%\#untrace 'x ...))])))

(define-syntax when
   (syntax-rules ()
      [(_ e0 e1 e2 ...)
       (if e0 (begin e1 e2 ...))]))

;;; case in Chez Scheme allows atomic keys

(define-syntax case
  (lambda (x)
    (syntax-case x ()
      ((_ e m1 m2 ...)
       (with-syntax
         ((body (let f ((clause (syntax m1)) (clauses (syntax (m2 ...))))
                  (if (null? clauses)
                      (syntax-case clause (else)
                        ((else e1 e2 ...) (syntax (begin e1 e2 ...)))
                        (((k ...) e1 e2 ...)
                         (syntax (if (#%memv t '(k ...)) (begin e1 e2 ...))))
                        ((k e1 e2 ...)
                         (syntax (if (#%memv t '(k)) (begin e1 e2 ...))))
                        (_ (syntax-error x)))
                      (with-syntax ((rest (f (car clauses) (cdr clauses))))
                        (syntax-case clause (else)
                          (((k ...) e1 e2 ...)
                           (syntax (if (#%memv t '(k ...))
                                       (begin e1 e2 ...)
                                       rest)))
                          ((k e1 e2 ...)
                           (syntax (if (#%memv t '(k))
                                       (begin e1 e2 ...)
                                       rest)))
                          (_ (syntax-error x))))))))
         (syntax (let ((t e)) body)))))))

(define-syntax delay
   (lambda (x)
      (syntax-case x ()
         ((delay exp)
          (syntax (#%\#make-promise (lambda () exp)))))))

(define-syntax recur
  (syntax-rules ()
    ((recur f ((i v) ...) e1 e2 ...)
     (let f ((i v) ...) e1 e2 ...))))

(define-syntax trace-recur
  (syntax-rules ()
    ((trace-recur f ((x v) ...) e1 e2 ...)
     (trace-let f ((x v) ...) e1 e2 ...))))

;; may want to get rid of this, but it's kind of convenient
(define-syntax identifier-syntax
  (lambda (x)
    (syntax-case x ()
      ((_ e)
       (syntax
         (lambda (x)
           (syntax-case x ()
             (id
              (identifier? (syntax id))
              (syntax e))
             ((id x (... ...))
;; how could this possibly fail if this proc is used as a transformer?
              (identifier? (syntax id))
              (syntax (e x (... ...)))))))))))

;; the identifier "id" is provided for use as a datum->syntax-object template.
;; could provide nicer error handling 

(define-syntax identifier-with-syntax
  (lambda (x)
    (syntax-case x ()
      ((_ clauses e)
       (with-syntax ((id (datum->syntax-object (syntax _) 'id)))
       (syntax
         (lambda (x)
           (syntax-case x ()
             (id
              (identifier? (syntax id))
              (with-syntax clauses (syntax e)))
             ((id x (... ...))
              (identifier? (syntax id))
              (with-syntax clauses (syntax (e x (... ...)))))))))))))
